from dataclasses import dataclass
from typing import Any, Dict, List, Optional, cast

from litellm.utils import cost_per_token, token_counter
from llama_index.callbacks.schema import CBEventType, EventPayload
from llama_index.callbacks.token_counting import TokenCountingEvent, TokenCountingHandler

from autollm.utils.logging import logger


@dataclass
class CostCalculatingEvent:
    prompt: str
    prompt_token_cost: int
    completion: str
    completion_token_cost: int
    total_token_cost: int = 0
    event_id: str = ""

    def __post_init__(self) -> None:
        self.total_token_cost = self.prompt_token_cost + self.completion_token_cost


def get_llm_token_counts(
        payload: Dict[str, Any], event_id: str = "", model: str = "gpt-3.5-turbo") -> TokenCountingEvent:
    from llama_index.llms import ChatMessage

    if EventPayload.PROMPT in payload:
        prompt = str(payload.get(EventPayload.PROMPT))
        completion = str(payload.get(EventPayload.COMPLETION))

        return TokenCountingEvent(
            event_id=event_id,
            prompt=prompt,
            prompt_token_count=token_counter(model=model, text=prompt),
            completion=completion,
            completion_token_count=token_counter(model=model, text=completion),
        )

    elif EventPayload.MESSAGES in payload:
        messages = cast(List[ChatMessage], payload.get(EventPayload.MESSAGES, []))
        messages_str = "\n".join([str(x) for x in messages])
        response = str(payload.get(EventPayload.RESPONSE))

        return TokenCountingEvent(
            event_id=event_id,
            prompt=messages_str,
            prompt_token_count=token_counter(model=model, text=messages_str),
            completion=response,
            completion_token_count=token_counter(model=model, text=response),
        )
    else:
        raise ValueError("Invalid payload! Need prompt and completion or messages and response.")


def get_llm_token_costs(
    latest_llm_token_count: TokenCountingEvent,
    event_id: str = "",
    model: str = "gpt-3.5-turbo",
) -> CostCalculatingEvent:
    """
    Calculate the cost of the LLM tokens.

    Args:
        latest_llm_token_count: The latest LLM token count.
        event_id: The event id.
        model: The model to use for calculating the cost.

    Returns:
        The latest LLM token cost. (CostCalculatingEvent)
    """
    prompt_tokens_cost_usd_dollar, completion_tokens_cost_usd_dollar = cost_per_token(
        model=model,
        prompt_tokens=latest_llm_token_count.prompt_token_count,
        completion_tokens=latest_llm_token_count.completion_token_count)
    return CostCalculatingEvent(
        event_id=event_id,
        prompt=latest_llm_token_count.prompt,
        prompt_token_cost=prompt_tokens_cost_usd_dollar,
        completion=latest_llm_token_count.completion,
        completion_token_cost=completion_tokens_cost_usd_dollar,
    )


class CostCalculatingHandler(TokenCountingHandler):
    """
    Callback handler for counting costs in LLM events. (Embeddings not supported yet)

    Parameters:
        model_name: The model to use for tokenizer to calculate the cost.
        event_starts_to_ignore: List of event types to ignore at the start of the event.
        event_ends_to_ignore: List of event types to ignore at the end of the event.
    """

    def __init__(
        self,
        model_name: str = "gpt-3.5-turbo",
        event_starts_to_ignore: Optional[List[CBEventType]] = None,
        event_ends_to_ignore: Optional[List[CBEventType]] = None,
        verbose: bool = False,
    ) -> None:
        self.llm_token_costs: List[CostCalculatingEvent] = []
        self.model_name = model_name

        super().__init__(
            event_starts_to_ignore=event_starts_to_ignore,
            event_ends_to_ignore=event_ends_to_ignore,
            verbose=verbose,
        )

    def on_event_end(
        self,
        event_type: CBEventType,
        payload: Optional[Dict[str, Any]] = None,
        event_id: str = "",
        **kwargs: Any,
    ) -> None:
        """Count the LLM or Embedding tokens as needed."""
        if (event_type == CBEventType.LLM and event_type not in self.event_ends_to_ignore and
                payload is not None):
            # token counts
            self.llm_token_counts.append(
                get_llm_token_counts(payload=payload, event_id=event_id, model=self.model_name))
            if self._verbose:
                logger.info(f"LLM Prompt Token Usage: {self.llm_token_counts[-1].prompt_token_count}")
                logger.info(f"LLM Completion Token Usage: {self.llm_token_counts[-1].completion_token_count}")

            # token costs
            self.llm_token_costs.append(
                get_llm_token_costs(
                    latest_llm_token_count=self.llm_token_counts[-1],
                    event_id=event_id,
                    model=self.model_name,
                ))
            if self._verbose:
                logger.info(f"LLM Latest Token Cost: ${self.llm_token_costs[-1].total_token_cost:.4f}")
                logger.info(f"LLM Total Token Cost: ${self.total_llm_token_cost:.4f}")

        elif (event_type == CBEventType.EMBEDDING and event_type not in self.event_ends_to_ignore and
              payload is not None):
            total_chunk_tokens = 0
            for chunk in payload.get(EventPayload.CHUNKS, []):
                self.embedding_token_counts.append(
                    TokenCountingEvent(
                        event_id=event_id,
                        prompt=chunk,
                        prompt_token_count=len(self.tokenizer(chunk)),
                        completion="",
                        completion_token_count=0,
                    ))
                total_chunk_tokens += self.embedding_token_counts[-1].total_token_count

            if self._verbose:
                logger.info(f"Embedding Token Usage: {total_chunk_tokens}")

    @property
    def total_llm_token_cost(self) -> int:
        """Get the current total LLM token cost."""
        return sum([x.total_token_cost for x in self.llm_token_costs])

    @property
    def prompt_llm_token_cost(self) -> int:
        """Get the current total LLM prompt token cost."""
        return sum([x.prompt_token_cost for x in self.llm_token_costs])

    @property
    def completion_llm_token_cost(self) -> int:
        """Get the current total LLM completion token cost."""
        return sum([x.completion_token_cost for x in self.llm_token_costs])



# Purpose: FastAPI documentation configuration.
# Metadata
title = "AutoLLM Query Engine"
description = """
This is a FastAPI service for AutoLLM's natural language query engine.
It's designed to query multiple and big documents and get the most relevant results.
"""
version = "0.0.1"
openapi_url = "/api/v1/openapi.json"
terms_of_service = "Local Deployment, All Rights Reserved."
tags_metadata = [
    {
        "name": "query",
        "description": "Operations related to querying the header-documents."
    },
]



from pathlib import Path

from autollm.utils.logging import logger


def clone_or_pull_repository(git_url: str, local_path: Path) -> None:
    """
    Clone a Git repository or pull latest changes if it already exists.

    Parameters:
        git_url (str): The URL of the Git repository.
        local_path (Path): The local path where the repository will be cloned or updated.
    """
    # Lazy import to avoid dependency on GitPython
    try:
        from git import InvalidGitRepositoryError, Repo
    except ImportError:
        logger.error(
            'GitPython is not installed. Please "pip install gitpython==3.1.37" to use this feature.')
        raise

    if local_path.exists():
        try:
            repo = Repo(str(local_path))
            repo.remotes.origin.pull()
        except InvalidGitRepositoryError:
            # The existing directory is not a valid git repo, clone anew
            Repo.clone_from(git_url, str(local_path))
    else:
        Repo.clone_from(git_url, str(local_path))



# Importing the necessary modules for testing and mocking

from fastapi.testclient import TestClient
from llama_index import Document, VectorStoreIndex
from llama_index.indices.query.base import BaseQueryEngine

from autollm.auto.fastapi_app import AutoFastAPI
from autollm.serve.utils import load_config_and_initialize_engines

# Mock the documents
documents = [Document.example()]

# Mock the yaml.safe_load to return a sample config
sample_config = {
    'summarize': {
        'system_prompt': 'You are a friendly chatbot that can summarize documents.'
    },
    'qa': {
        'system_prompt': 'You are a friendly chatbot that can answer questions.'
    }
}


def test_load_config_and_initialize_engines():
    # Mock the load_config_and_initialize_engines function with the sample config and documents
    query_engines = load_config_and_initialize_engines('tests/config.yaml', documents=documents)

    # Validate the type and configuration of each query engine
    for task in sample_config.keys():
        assert isinstance(query_engines[task], BaseQueryEngine)


def test_auto_fastapi_from_config():
    app = AutoFastAPI.from_config(config_file_path='tests/config.yaml', documents=documents)

    # Validate presence of the query endpoint
    assert any(route.path == "/query" for route in app.routes)


def test_auto_fastapi_from_query_engine():
    from llama_index import VectorStoreIndex
    index = VectorStoreIndex.from_documents(documents=documents)
    query_engine = index.as_query_engine()

    # Create the FastAPI app from the query engine
    app = AutoFastAPI.from_query_engine(query_engine=query_engine)

    # Validate presence of the query endpoint
    assert any(route.path == "/query" for route in app.routes)


def test_query_endpoint_from_config():
    # Create the FastAPI app with test configuration
    app = AutoFastAPI.from_config('tests/config.yaml', documents=documents)
    client = TestClient(app)

    # Test with a valid task (changed "task1" to "summarize")
    response = client.post("/query", json={"task": "summarize", "user_query": "test query"})
    assert response.status_code == 200

    # Test with an invalid task
    response = client.post("/query", json={"task": "invalid_task", "user_query": "test query"})
    assert response.status_code == 400


def test_query_endpoint_from_query_engine():
    # Create llama-index query engine
    index = VectorStoreIndex.from_documents(documents=documents)
    query_engine = index.as_query_engine()

    # Create the FastAPI app from the query engine
    app = AutoFastAPI.from_query_engine(query_engine=query_engine)
    client = TestClient(app)

    # Test with a user query
    response = client.post("/query", json={"user_query": "why so serious?"})
    assert response.status_code == 200



from typing import List

import requests
from bs4 import BeautifulSoup
from llama_index.schema import Document

from autollm.utils.constants import WEBPAGE_READER_TIMEOUT
from autollm.utils.logging import logger

# Constants defined outside the class
SELECTORS = [
    "article.bd-article",
    'article[role="main"]',
    "div.md-content",
    'div[role="main"]',
    "div.container",
    "div.section",
    "article",
    "main",
]

IGNORED_TAGS = [
    "nav",
    "aside",
    "form",
    "header",
    "noscript",
    "svg",
    "canvas",
    "footer",
    "script",
    "style",
]


class WebPageReader:
    """A class for reading and processing the content of a single web page."""

    def load_data(self, url: str) -> List[Document]:
        """
        Reads a web page from the provided URL, extracts content using predefined selectors, removes all
        ignored tags, and returns a list of Document objects with the processed content.

        Parameters:
            url (str): The URL of the web page to read.

        Returns:
            List[Document]: A list containing Document objects with the processed content and its metadata.
        """
        response = requests.get(url, timeout=WEBPAGE_READER_TIMEOUT)
        if response.status_code != 200:
            logger.info(f"Failed to fetch the website: {response.status_code}")
            return []

        soup = BeautifulSoup(response.content, "html.parser")

        for selector in SELECTORS:
            element = soup.select_one(selector)
            if element:
                content = element.prettify()
                break
        else:
            logger.info(f"Failed to find any element for URL: {url}")
            content = soup.get_text()

        soup = BeautifulSoup(content, "html.parser")
        for tag in soup(IGNORED_TAGS):
            tag.decompose()

        content = " ".join(soup.stripped_strings)
        document = Document(id_=url, text=content, metadata={"url": url})
        logger.info(f"Processed URL: {url}")

        return [document]


# Example of usage:
# reader = WebPageReader()
# documents = reader.load_data('http://example.com/some-page')



"""LanceDB vector store with cloud storage support."""
import os
from typing import Any, Optional

from dotenv import load_dotenv
from llama_index.schema import NodeRelationship, RelatedNodeInfo, TextNode
from llama_index.vector_stores import LanceDBVectorStore as LanceDBVectorStoreBase
from llama_index.vector_stores.lancedb import _to_lance_filter, _to_llama_similarities
from llama_index.vector_stores.types import VectorStoreQuery, VectorStoreQueryResult
from pandas import DataFrame

load_dotenv()


class LanceDBVectorStore(LanceDBVectorStoreBase):
    """Advanced LanceDB Vector Store supporting cloud storage and prefiltering."""
    from lancedb.query import LanceQueryBuilder
    from lancedb.table import Table

    def __init__(
        self,
        uri: str,
        table_name: str = "vectors",
        nprobes: int = 20,
        refine_factor: Optional[int] = None,
        api_key: Optional[str] = None,
        region: Optional[str] = None,
        **kwargs: Any,
    ) -> None:
        """Init params."""
        self._setup_connection(uri, api_key, region)
        self.uri = uri
        self.table_name = table_name
        self.nprobes = nprobes
        self.refine_factor = refine_factor
        self.api_key = api_key
        self.region = region

    def _setup_connection(self, uri: str, api_key: Optional[str] = None, region: Optional[str] = None):
        """Establishes a robust connection to LanceDB."""
        api_key = api_key or os.getenv('LANCEDB_API_KEY')
        region = region or os.getenv('LANCEDB_REGION')

        import_err_msg = "`lancedb` package not found, please run `pip install lancedb`"
        try:
            import lancedb
        except ImportError:
            raise ImportError(import_err_msg)

        if api_key and region:
            self.connection = lancedb.connect(uri, api_key=api_key, region=region)
        else:
            self.connection = lancedb.connect(uri)

    def query(
        self,
        query: VectorStoreQuery,
        **kwargs: Any,
    ) -> VectorStoreQueryResult:
        """Enhanced query method to support prefiltering in LanceDB queries."""
        table = self.connection.open_table(self.table_name)
        lance_query = self._prepare_lance_query(query, table, **kwargs)

        results = lance_query.to_df()
        return self._construct_query_result(results)

    def _prepare_lance_query(self, query: VectorStoreQuery, table: Table, **kwargs) -> LanceQueryBuilder:
        """Prepares the LanceDB query considering prefiltering and additional parameters."""
        if query.filters is not None:
            if "where" in kwargs:
                raise ValueError(
                    "Cannot specify filter via both query and kwargs. "
                    "Use kwargs only for lancedb specific items that are "
                    "not supported via the generic query interface.")
            where = _to_lance_filter(query.filters)
        else:
            where = kwargs.pop("where", None)
        prefilter = kwargs.pop("prefilter", False)

        table = self.connection.open_table(self.table_name)
        lance_query = (
            table.search(query.query_embedding).limit(query.similarity_top_k).where(
                where, prefilter=prefilter).nprobes(self.nprobes))

        if self.refine_factor is not None:
            lance_query.refine_factor(self.refine_factor)

        return lance_query

    def _construct_query_result(self, results: DataFrame) -> VectorStoreQueryResult:
        """Constructs a VectorStoreQueryResult from a LanceDB query result."""
        nodes = []

        for _, row in results.iterrows():
            node = TextNode(
                text=row.get('text', ''),  # ensure text is a string
                id_=row['id'],
                relationships={
                    NodeRelationship.SOURCE: RelatedNodeInfo(node_id=row['doc_id']),
                })
            nodes.append(node)

        return VectorStoreQueryResult(
            nodes=nodes,
            similarities=_to_llama_similarities(results),
            ids=results["id"].tolist(),
        )



"""AutoLLM: A Base Package for Large Language Model Applications.

This package provides automated integrations with leading large language models
and vector databases, along with various utility functions.
"""

__version__ = '0.1.10'
__author__ = 'safevideo'
__license__ = 'AGPL-3.0'

from autollm.auto.embedding import AutoEmbedding
from autollm.auto.fastapi_app import AutoFastAPI
from autollm.auto.llm import AutoLiteLLM
from autollm.auto.query_engine import AutoQueryEngine
from autollm.auto.service_context import AutoServiceContext
from autollm.auto.vector_store_index import AutoVectorStoreIndex
from autollm.utils.document_reading import (
    read_files_as_documents,
    read_github_repo_as_documents,
    read_webpage_as_documents,
    read_website_as_documents,
)

__all__ = [
    'AutoLiteLLM', 'AutoServiceContext', 'AutoVectorStoreIndex', 'AutoQueryEngine', 'AutoFastAPI',
    'AutoEmbedding', 'read_files_as_documents', 'read_github_repo_as_documents', 'read_webpage_as_documents',
    'read_website_as_documents'
]



from typing import Optional, Union

from llama_index import ServiceContext
from llama_index.callbacks import CallbackManager
from llama_index.embeddings.utils import EmbedType
from llama_index.extractors import (
    EntityExtractor,
    KeywordExtractor,
    QuestionsAnsweredExtractor,
    SummaryExtractor,
    TitleExtractor,
)
from llama_index.llms.utils import LLMType
from llama_index.prompts import PromptTemplate
from llama_index.prompts.base import BasePromptTemplate
from llama_index.text_splitter import SentenceSplitter

from autollm.callbacks.cost_calculating import CostCalculatingHandler
from autollm.utils.llm_utils import set_default_prompt_template


class AutoServiceContext:
    """AutoServiceContext extends the functionality of LlamaIndex's ServiceContext to include token
    counting.
    """

    @staticmethod
    def from_defaults(
            llm: Optional[LLMType] = "default",
            embed_model: Optional[EmbedType] = "default",
            system_prompt: str = None,
            query_wrapper_prompt: Union[str, BasePromptTemplate] = None,
            enable_cost_calculator: bool = False,
            chunk_size: Optional[int] = 512,
            chunk_overlap: Optional[int] = 100,
            context_window: Optional[int] = None,
            enable_title_extractor: bool = False,
            enable_summary_extractor: bool = False,
            enable_qa_extractor: bool = False,
            enable_keyword_extractor: bool = False,
            enable_entity_extractor: bool = False,
            **kwargs) -> ServiceContext:
        """
        Create a ServiceContext with default parameters with extended enable_token_counting functionality. If
        enable_token_counting is True, tracks the number of tokens used by the LLM for each query.

        Parameters:
            llm (LLM): The LLM to use for the query engine. Defaults to gpt-3.5-turbo.
            embed_model (BaseEmbedding): The embedding model to use for the query engine. Defaults to OpenAIEmbedding.
            system_prompt (str): The system prompt to use for the query engine.
            query_wrapper_prompt (Union[str, BasePromptTemplate]): The query wrapper prompt to use for the query engine.
            enable_cost_calculator (bool): Flag to enable cost calculator logging.
            chunk_size (int): The token chunk size for each chunk.
            chunk_overlap (int): The token overlap between each chunk.
            context_window (int): The maximum context size that will get sent to the LLM.
            enable_title_extractor (bool): Flag to enable title extractor.
            enable_summary_extractor (bool): Flag to enable summary extractor.
            enable_qa_extractor (bool): Flag to enable question answering extractor.
            enable_keyword_extractor (bool): Flag to enable keyword extractor.
            enable_entity_extractor (bool): Flag to enable entity extractor.
            **kwargs: Arbitrary keyword arguments.

        Returns:
            ServiceContext: The initialized ServiceContext from default parameters with extra token counting functionality.
        """
        if not system_prompt and not query_wrapper_prompt:
            system_prompt, query_wrapper_prompt = set_default_prompt_template()
        # Convert query_wrapper_prompt to PromptTemplate if it is a string
        if isinstance(query_wrapper_prompt, str):
            query_wrapper_prompt = PromptTemplate(template=query_wrapper_prompt)

        callback_manager: CallbackManager = kwargs.get('callback_manager', CallbackManager())
        kwargs.pop(
            'callback_manager', None)  # Make sure callback_manager is not passed to ServiceContext twice

        if enable_cost_calculator:
            llm_model_name = llm.metadata.model_name if not "default" else "gpt-3.5-turbo"
            callback_manager.add_handler(CostCalculatingHandler(model_name=llm_model_name, verbose=True))

        sentence_splitter = SentenceSplitter(chunk_size=chunk_size, chunk_overlap=chunk_overlap)
        transformations = [sentence_splitter]
        if enable_entity_extractor:
            transformations.append(EntityExtractor())
        if enable_keyword_extractor:
            transformations.append(KeywordExtractor(llm=llm, keywords=5))
        if enable_summary_extractor:
            transformations.append(SummaryExtractor(llm=llm, summaries=["prev", "self"]))
        if enable_title_extractor:
            transformations.append(TitleExtractor(llm=llm, nodes=5))
        if enable_qa_extractor:
            transformations.append(QuestionsAnsweredExtractor(llm=llm, questions=5))

        service_context = ServiceContext.from_defaults(
            llm=llm,
            embed_model=embed_model,
            transformations=transformations,
            system_prompt=system_prompt,
            query_wrapper_prompt=query_wrapper_prompt,
            chunk_size=chunk_size,
            chunk_overlap=chunk_overlap,
            context_window=context_window,
            callback_manager=callback_manager,
            **kwargs)

        return service_context



from llama_index import Document, ServiceContext, VectorStoreIndex
from llama_index.llms.base import BaseLLM
from llama_index.query_engine import BaseQueryEngine

from autollm.auto.llm import AutoLiteLLM


def test_auto_lite_llm():
    document = Document.example()

    llm = AutoLiteLLM.from_defaults(model="gpt-3.5-turbo")

    # Check if the llm is an instance of LLM
    assert isinstance(llm, BaseLLM)

    service_context = ServiceContext.from_defaults(llm=llm)

    index = VectorStoreIndex.from_documents(documents=[document], service_context=service_context)

    query_engine = index.as_query_engine()

    # Check if the query_engine is an instance of BaseQueryEngine
    assert isinstance(query_engine, BaseQueryEngine)



# autollm/utils/logging.py

import logging

# Configure the logger
logger = logging.getLogger('autollm')
logger.setLevel(logging.INFO)

# Create a console handler with a specific format
ch = logging.StreamHandler()
ch.setLevel(logging.INFO)
formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
ch.setFormatter(formatter)

# Add the handler to the logger
logger.addHandler(ch)

# Ensure that log messages propagate up to the root logger
logger.propagate = False



from typing import Optional, Sequence, Union

from llama_index import Document, ServiceContext, VectorStoreIndex
from llama_index.embeddings.utils import EmbedType
from llama_index.indices.query.base import BaseQueryEngine
from llama_index.prompts.base import BasePromptTemplate, PromptTemplate
from llama_index.prompts.prompt_type import PromptType
from llama_index.response_synthesizers import get_response_synthesizer
from llama_index.schema import BaseNode

from autollm.auto.embedding import AutoEmbedding
from autollm.auto.llm import AutoLiteLLM
from autollm.auto.service_context import AutoServiceContext
from autollm.auto.vector_store_index import AutoVectorStoreIndex
from autollm.utils.env_utils import load_config_and_dotenv


def create_query_engine(
        documents: Optional[Sequence[Document]] = None,
        nodes: Optional[Sequence[BaseNode]] = None,
        # llm_params
        llm_model: str = "gpt-3.5-turbo",
        llm_max_tokens: Optional[int] = 256,
        llm_temperature: float = 0.1,
        llm_api_base: Optional[str] = None,
        # service_context_params
        system_prompt: str = None,
        query_wrapper_prompt: Union[str, BasePromptTemplate] = None,
        enable_cost_calculator: bool = True,
        embed_model: Optional[str] = "text-embedding-ada-002",
        chunk_size: Optional[int] = 512,
        chunk_overlap: Optional[int] = 100,
        context_window: Optional[int] = None,
        enable_title_extractor: bool = False,
        enable_summary_extractor: bool = False,
        enable_qa_extractor: bool = False,
        enable_keyword_extractor: bool = False,
        enable_entity_extractor: bool = False,
        # query_engine_params
        similarity_top_k: int = 6,
        response_mode: str = "compact",
        refine_prompt: str = None,
        structured_answer_filtering: bool = False,
        # vector_store_params
        vector_store_type: str = "LanceDBVectorStore",
        lancedb_uri: str = "./.lancedb",
        lancedb_table_name: str = "vectors",
        use_async: bool = True,
        exist_ok: bool = False,
        overwrite_existing: bool = False,
        **vector_store_kwargs) -> BaseQueryEngine:
    """
    Create a query engine from parameters.

    Parameters:
        documents (Sequence[Document]): Sequence of llama_index.Document instances.
        nodes (Sequence[BaseNode]): Sequence of llama_index.BaseNode instances.
        llm_model (str): The LLM model to use for the query engine.
        llm_max_tokens (int): The maximum number of tokens to be generated as LLM output.
        llm_temperature (float): The temperature to use for the LLM.
        llm_api_base (str): The API base to use for the LLM.
        system_prompt (str): The system prompt to use for the query engine.
        query_wrapper_prompt (Union[str, BasePromptTemplate]): The query wrapper prompt to use for the query engine.
        enable_cost_calculator (bool): Flag to enable cost calculator logging.
        embed_model (Union[str, EmbedType]): The embedding model to use for generating embeddings. "default" for OpenAI,
                                            "local" for HuggingFace or use full identifier (e.g., local:intfloat/multilingual-e5-large)
        chunk_size (int): The token chunk size for each chunk.
        chunk_overlap (int): The token overlap between each chunk.
        context_window (int): The maximum context size that will get sent to the LLM.
        enable_title_extractor (bool): Flag to enable title extractor.
        enable_summary_extractor (bool): Flag to enable summary extractor.
        enable_qa_extractor (bool): Flag to enable question answering extractor.
        enable_keyword_extractor (bool): Flag to enable keyword extractor.
        enable_entity_extractor (bool): Flag to enable entity extractor.
        similarity_top_k (int): The number of similar documents to return.
        response_mode (str): The response mode to use for the query engine.
        refine_prompt (str): The refine prompt to use for the query engine.
        vector_store_type (str): The vector store type to use for the query engine.
        lancedb_uri (str): The URI to use for the LanceDB vector store.
        lancedb_table_name (str): The table name to use for the LanceDB vector store.

    Returns:
        A llama_index.BaseQueryEngine instance.
    """
    llm = AutoLiteLLM.from_defaults(
        model=llm_model, api_base=llm_api_base, max_tokens=llm_max_tokens, temperature=llm_temperature)

    embedding = AutoEmbedding(model=embed_model)

    service_context = AutoServiceContext.from_defaults(
        llm=llm,
        embed_model=embedding,
        system_prompt=system_prompt,
        query_wrapper_prompt=query_wrapper_prompt,
        enable_cost_calculator=enable_cost_calculator,
        chunk_size=chunk_size,
        chunk_overlap=chunk_overlap,
        context_window=context_window,
        enable_title_extractor=enable_title_extractor,
        enable_summary_extractor=enable_summary_extractor,
        enable_qa_extractor=enable_qa_extractor,
        enable_keyword_extractor=enable_keyword_extractor,
        enable_entity_extractor=enable_entity_extractor,
    )
    vector_store_index = AutoVectorStoreIndex.from_defaults(
        vector_store_type=vector_store_type,
        lancedb_uri=lancedb_uri,
        lancedb_table_name=lancedb_table_name,
        use_async=use_async,
        documents=documents,
        nodes=nodes,
        service_context=service_context,
        exist_ok=exist_ok,
        overwrite_existing=overwrite_existing,
        **vector_store_kwargs)
    if refine_prompt is not None:
        refine_prompt_template = PromptTemplate(refine_prompt, prompt_type=PromptType.REFINE)
    else:
        refine_prompt_template = None

    # Convert query_wrapper_prompt to PromptTemplate if it is a string
    if isinstance(query_wrapper_prompt, str):
        query_wrapper_prompt = PromptTemplate(template=query_wrapper_prompt)
    response_synthesizer = get_response_synthesizer(
        service_context=service_context,
        text_qa_template=query_wrapper_prompt,
        refine_template=refine_prompt_template,
        response_mode=response_mode,
        structured_answer_filtering=structured_answer_filtering)

    return vector_store_index.as_query_engine(
        similarity_top_k=similarity_top_k, response_synthesizer=response_synthesizer)


class AutoQueryEngine:
    """
    AutoQueryEngine for query execution and optionally logging the query cost.

    ```python
    from autollm.auto.query_engine import AutoQueryEngine

    # Create an AutoQueryEngine from a config file and .env file
    query_engine = AutoQueryEngine.from_config("config.yaml", ".env")

    # Create an AutoQueryEngine from a vector store index and service context
    query_engine = AutoQueryEngine.from_instances(vector_store_index, service_context)

    # Create an AutoQueryEngine from defaults
    query_engine = AutoQueryEngine.from_defaults(
        documents=documents,
        # llm_params
        llm_model="gpt-3.5-turbo",
        llm_api_base=None,
        llm_max_tokens=None,
        llm_temperature=0.1,
        # service_context_params
        system_prompt=None,
        query_wrapper_prompt=None,
        enable_cost_calculator=True,
        embed_model="text-embedding-ada-002",
        chunk_size=512,
        chunk_overlap=None,
        context_window=None,
        # query_engine_params
        similarity_top_k=6,
        # vector_store_params
        vector_store_type="LanceDBVectorStore",
        lancedb_uri="./.lancedb",
        lancedb_table_name="vectors",
        **vector_store_kwargs)
    )
    ```
    """

    @staticmethod
    def from_instances(
            vector_store_index: VectorStoreIndex, service_context: ServiceContext,
            **kwargs) -> BaseQueryEngine:
        """
        Create an AutoQueryEngine from a vector store index and a service context.

        Parameters:
            vector_store_index: llama_index.VectorStoreIndex instance.
            service_context: llama_index.ServiceContext instance.
            **kwargs: Keyword arguments for the query engine.

        Returns:
            A llama_index.BaseQueryEngine instance.
        """

        return vector_store_index.as_query_engine(service_context=service_context, **kwargs)

    @staticmethod
    def from_defaults(
            documents: Optional[Sequence[Document]] = None,
            nodes: Optional[Sequence[BaseNode]] = None,
            # llm_params
            llm_model: Optional[str] = "gpt-3.5-turbo",
            llm_api_base: Optional[str] = None,
            llm_max_tokens: Optional[int] = None,
            llm_temperature: Optional[float] = 0.1,
            # service_context_params
            system_prompt: str = None,
            query_wrapper_prompt: Union[str, BasePromptTemplate] = None,
            enable_cost_calculator: bool = True,
            embed_model: Optional[str] = "text-embedding-ada-002",
            chunk_size: Optional[int] = 512,
            chunk_overlap: Optional[int] = 200,
            context_window: Optional[int] = None,
            # query_engine_params
            similarity_top_k: int = 6,
            response_mode: str = "compact",
            refine_prompt: str = None,
            structured_answer_filtering: bool = False,
            # vector_store_params
            vector_store_type: str = "LanceDBVectorStore",
            lancedb_uri: str = "./.lancedb",
            lancedb_table_name: str = "vectors",
            use_async: bool = True,
            exist_ok: bool = False,
            overwrite_existing: bool = False,
            **vector_store_kwargs) -> BaseQueryEngine:
        """
        Create an AutoQueryEngine from default parameters.

        Parameters:
            documents (Sequence[Document]): Sequence of llama_index.Document instances.
            nodes (Sequence[BaseNode]): Sequence of llama_index.BaseNode instances.
            llm_model (str): The LLM model to use for the query engine.
            llm_max_tokens (int): The maximum number of tokens to be generated as LLM output.
            llm_temperature (float): The temperature to use for the LLM.
            llm_api_base (str): The API base to use for the LLM.
            system_prompt (str): The system prompt to use for the query engine.
            query_wrapper_prompt (Union[str, BasePromptTemplate]): The query wrapper prompt to use for the query engine.
            enable_cost_calculator (bool): Flag to enable cost calculator logging.
            embed_model (Union[str, EmbedType]): The embedding model to use for generating embeddings.
            chunk_size (int): The token chunk size for each chunk.
            chunk_overlap (int): The token overlap between each chunk.
            context_window (int): The maximum context size that will get sent to the LLM.
            enable_title_extractor (bool): Flag to enable title extractor.
            enable_summary_extractor (bool): Flag to enable summary extractor.
            enable_qa_extractor (bool): Flag to enable question answering extractor.
            enable_keyword_extractor (bool): Flag to enable keyword extractor.
            enable_entity_extractor (bool): Flag to enable entity extractor.
            similarity_top_k (int): The number of similar documents to return.
            response_mode (str): The response mode to use for the query engine.
            refine_prompt (str): The refine prompt to use for the query engine.
            vector_store_type (str): The vector store type to use for the query engine.
            lancedb_uri (str): The URI to use for the LanceDB vector store.
            lancedb_table_name (str): The table name to use for the LanceDB vector store.
            exist_ok (bool): Flag to allow overwriting an existing vector store.
            overwrite_existing (bool): Flag to allow overwriting an existing vector store.

            Returns:
                A llama_index.BaseQueryEngine instance.
        """

        return create_query_engine(
            documents=documents,
            nodes=nodes,
            # llm_params
            llm_model=llm_model,
            llm_api_base=llm_api_base,
            llm_max_tokens=llm_max_tokens,
            llm_temperature=llm_temperature,
            # service_context_params
            system_prompt=system_prompt,
            query_wrapper_prompt=query_wrapper_prompt,
            enable_cost_calculator=enable_cost_calculator,
            embed_model=embed_model,
            chunk_size=chunk_size,
            chunk_overlap=chunk_overlap,
            context_window=context_window,
            # query_engine_params
            similarity_top_k=similarity_top_k,
            response_mode=response_mode,
            refine_prompt=refine_prompt,
            structured_answer_filtering=structured_answer_filtering,
            # vector_store_params
            vector_store_type=vector_store_type,
            lancedb_uri=lancedb_uri,
            lancedb_table_name=lancedb_table_name,
            use_async=use_async,
            exist_ok=exist_ok,
            overwrite_existing=overwrite_existing,
            **vector_store_kwargs)

    @staticmethod
    def from_config(
            config_file_path: str,
            env_file_path: str = None,
            documents: Sequence[Document] = None,
            nodes: Optional[Sequence[BaseNode]] = None) -> BaseQueryEngine:
        """
        Create an AutoQueryEngine from a config file and optionally a .env file.

        Parameters:
            config_file_path (str): Path to the YAML configuration file.
            env_file_path (str): Path to the .env file.
            documents (Sequence[Document]): Sequence of llama_index.Document instances.

        Returns:
            A llama_index.BaseQueryEngine instance.
        """

        config = load_config_and_dotenv(config_file_path, env_file_path)
        # Get the first task configuration
        config = config['tasks'][0]

        return create_query_engine(
            documents=documents,
            nodes=nodes,
            **config,
        )



import xml.etree.ElementTree as ET
from typing import List, Optional
from urllib.parse import urljoin, urlparse

import requests
from bs4 import BeautifulSoup
from llama_index.schema import Document
from tqdm import tqdm

from autollm.utils.constants import WEBPAGE_READER_TIMEOUT
from autollm.utils.logging import logger
from autollm.utils.webpage_reader import WebPageReader


class WebSiteReader:

    def __init__(self):
        self.visited_links = set()

    def _fetch_and_parse_sitemap(
            self,
            sitemap_url: str,
            include_filter_str: Optional[str] = None,
            exclude_filter_str: Optional[str] = None) -> List[str]:
        """Fetches and parses the sitemap, returning URLs."""
        try:
            response = requests.get(sitemap_url)
            response.raise_for_status()
            sitemap_content = response.text
        except requests.RequestException as e:
            logger.error(f"Error fetching sitemap: {e}")
            return []

        return self._extract_urls_from_sitemap(sitemap_content, include_filter_str, exclude_filter_str)

    def _extract_urls_from_sitemap(
            self, sitemap_content: str, include_filter_str: Optional[str],
            exclude_filter_str: Optional[str]) -> List[str]:
        """Extracts URLs from sitemap content."""
        sitemap = ET.fromstring(sitemap_content)
        urls = []

        for url in sitemap.findall(".//{http://www.sitemaps.org/schemas/sitemap/0.9}url"):
            location = url.find("{http://www.sitemaps.org/schemas/sitemap/0.9}loc").text
            if (include_filter_str and
                    include_filter_str not in location) or (exclude_filter_str and
                                                            exclude_filter_str in location):
                continue
            urls.append(location)

        return urls

    def _get_child_links_recursive(self, url):
        parsed_url = urlparse(url)
        base_url = f"{parsed_url.scheme}://{parsed_url.netloc}"
        current_path = parsed_url.path

        response = requests.get(url, timeout=WEBPAGE_READER_TIMEOUT)  # timeout in seconds
        if response.status_code != 200:
            logger.warning(f"Failed to fetch the website: {response.status_code}")
            return

        soup = BeautifulSoup(response.text, "html.parser")
        all_links = [link.get("href") for link in soup.find_all("a")]

        # Normalize links and filter out external links and anchors
        child_links = set()
        for link in all_links:
            # Skip any None or empty hrefs, and anchors
            if not link or link.startswith('#') or link == current_path:
                continue
            # Convert relative links to absolute
            full_link = urljoin(base_url, link)
            # Add to set if the link is internal
            if urlparse(full_link).netloc == parsed_url.netloc:
                child_links.add(full_link)

        # Process each child link
        for link in child_links:
            if link not in self.visited_links:
                self.visited_links.add(link)
                self._get_child_links_recursive(link)

    def _get_all_urls(self, url):
        self.visited_links = set()
        self._get_child_links_recursive(url)
        urls = [link for link in self.visited_links if urlparse(link).netloc == urlparse(url).netloc]
        return urls

    def load_data(
            self,
            parent_url: str = None,
            sitemap_url: str = None,
            include_filter_str: str = None,
            exclude_filter_str: str = None) -> List[Document]:
        """Loads data from either a standard URL or a sitemap URL."""
        if sitemap_url:
            logger.info(f"Fetching and parsing sitemap {sitemap_url}..")
            all_urls = self._fetch_and_parse_sitemap(sitemap_url, include_filter_str, exclude_filter_str)
        elif parent_url:
            logger.info(f"Parsing child pages of {parent_url}..")
            all_urls = self._get_child_links_recursive(parent_url)
        else:
            raise ValueError("Either sitemap_url or parent_url must be provided.")

        web_reader = WebPageReader()
        documents = []

        logger.info(f"Total URLs to process: {len(all_urls)}")

        for u in tqdm(all_urls, desc="Processing URLs"):
            if u not in self.visited_links:
                self.visited_links.add(u)
                documents.extend(web_reader.load_data(u))

        return documents



from typing import List

from llama_index.readers.base import BaseReader
from llama_index.schema import Document

from autollm.utils.logging import logger


class LangchainPDFReader(BaseReader):
    """Custom PDF reader that uses langchain's PDFMinerLoader."""

    def __init__(self, extract_images: bool = False) -> None:
        """Initialize the reader."""
        self.extract_images = extract_images

    def load_data(self, file_path: str, extra_info: dict = None) -> List[Document]:
        """Load data from a PDF file using langchain's PDFMinerLoader."""
        from langchain.document_loaders import PDFMinerLoader

        # Convert the PosixPath object to a string before passing it to PDFMinerLoader
        loader = PDFMinerLoader(str(file_path), extract_images=self.extract_images)

        langchain_documents = loader.load()  # This returns a list of langchain Document objects

        # Convert langchain documents into llama-index documents
        documents = []
        for langchain_document in langchain_documents:
            # Create a llama-index document for each langchain document
            doc = Document.from_langchain_format(langchain_document)

            # If there's extra info, we can add it to the Document's metadata
            if extra_info is not None:
                doc.metadata.update(extra_info)

            documents.append(doc)

        return documents



from llama_index import Document, VectorStoreIndex
from llama_index.query_engine import BaseQueryEngine

from autollm.auto.query_engine import AutoQueryEngine

documents = [Document.example()]


def test_auto_query_engine_from_defaults():
    vector_store_type = "SimpleVectorStore"
    query_engine = AutoQueryEngine.from_defaults(documents=documents, vector_store_type=vector_store_type)

    # Check if the query_engine is an instance of BaseQueryEngine
    assert isinstance(query_engine, BaseQueryEngine)


def test_auto_query_engine_from_instances():
    vector_store_index = VectorStoreIndex.from_documents(documents=documents)

    query_engine = AutoQueryEngine.from_instances(vector_store_index=vector_store_index, service_context=None)

    # Check if the query_engine is an instance of BaseQueryEngine
    assert isinstance(query_engine, BaseQueryEngine)


def test_auto_query_engine_from_config():
    query_engine = AutoQueryEngine.from_config(config_file_path="tests/config.yaml", documents=documents)

    # Check if the query_engine is an instance of BaseQueryEngine
    assert isinstance(query_engine, BaseQueryEngine)



"""
Markdown Reader.

A parser for md files.
"""
import re
import uuid
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple, cast

from llama_index.readers.base import BaseReader
from llama_index.schema import Document


class MarkdownReader(BaseReader):
    """
    Markdown parser.

    Extract text from markdown files. Returns dictionary with keys as headers and values as the text between
    headers.
    """

    def __init__(
        self,
        *args: Any,
        remove_hyperlinks: bool = True,
        remove_images: bool = True,
        read_as_single_doc: bool = False,
        **kwargs: Any,
    ) -> None:
        """Init params."""
        super().__init__(*args, **kwargs)
        self._remove_hyperlinks = remove_hyperlinks
        self._remove_images = remove_images
        self._read_as_single_doc = read_as_single_doc

    def markdown_to_tups(self, markdown_text: str) -> List[Tuple[Optional[str], str]]:
        """
        Convert a markdown file to a dictionary.

        The keys are the headers and the values are the text under each header.
        """
        markdown_tups: List[Tuple[Optional[str], str]] = []
        lines = markdown_text.split('\n')

        current_header = None
        current_text = ''

        for line in lines:
            header_match = re.match(r'^#+\s', line)
            if header_match:
                if current_header is not None:
                    if current_text == '' or None:
                        continue
                    markdown_tups.append((current_header, current_text))

                current_header = line
                current_text = ''
            else:
                current_text += line + '\n'
        markdown_tups.append((current_header, current_text))

        if current_header is not None:
            # pass linting, assert keys are defined
            markdown_tups = [(re.sub(r'#', '', cast(str, key)).strip(), re.sub(r'<.*?>', '', value))
                             for key, value in markdown_tups]
        else:
            markdown_tups = [(key, re.sub('<.*?>', '', value)) for key, value in markdown_tups]

        return markdown_tups

    def remove_images(self, content: str) -> str:
        """Get a dictionary of a markdown file from its path."""
        pattern = r'!{1}\[\[(.*)\]\]'
        content = re.sub(pattern, '', content)
        return content

    def remove_hyperlinks(self, content: str) -> str:
        """Get a dictionary of a markdown file from its path."""
        pattern = r'\[(.*?)\]\((.*?)\)'
        content = re.sub(pattern, r'\1', content)
        return content

    def parse_tups(self,
                   filepath: Path,
                   content: Optional[str] = None,
                   errors: str = 'ignore') -> List[Tuple[Optional[str], str]]:
        """
        Parse file into tuples.

        If content is provided, use that instead of reading from file.
        """
        if content is None:
            with open(filepath, encoding='utf-8') as f:
                content = f.read()
        if self._remove_hyperlinks:
            content = self.remove_hyperlinks(content)
        if self._remove_images:
            content = self.remove_images(content)
        markdown_tups = self.markdown_to_tups(content)
        return markdown_tups

    def _load_data_as_sections(
        self,
        file: Path,
        extra_info: Optional[Dict] = None,
        content: Optional[str] = None,
    ) -> List[Document]:
        """
        Parse file into string. If content is provided, use that instead of reading from file.

        Parameters:
            file (Path): The path to the markdown file.
            extra_info (Optional[Dict]): Additional metadata to include.
            content (Optional[str]): Content to use instead of reading from file.

        Returns:
            List[Document]: List of Document objects representing header-docs.
        """
        tups = self.parse_tups(file, content=content)
        results = []

        for header, value in tups:
            # Generating doc_id
            doc_id = f"{str(file)}_{header.replace(' ', '_')}" if header else str(uuid.uuid4())

            # Creating Document object
            results.append(
                Document(
                    id_=doc_id,  # Set the id_
                    text=f'\n\n{header}\n{value}' if header else value,
                    metadata=extra_info or {}))

        return results

    def _load_data_as_single_doc(
        self,
        file: Path,
        extra_info: Optional[Dict] = None,
        content: Optional[str] = None,
    ) -> List[Document]:
        """
        Parse file into string. If content is provided, use that instead of reading from file.

        Parameters:
            file (Path): The path to the markdown file.
            extra_info (Optional[Dict]): Additional metadata to include.
            content (Optional[str]): Content to use instead of reading from file.

        Returns:
            List[Document]: List of a single Document object representing markdown text.
        """
        # Reading entire markdown as a single document
        with open(file, encoding='utf-8') as f:
            content = f.read()
        if self._remove_hyperlinks:
            content = self.remove_hyperlinks(content)
        if self._remove_images:
            content = self.remove_images(content)

        return [Document(text=content, metadata=extra_info)]

    def load_data(self,
                  file: Path,
                  extra_info: Optional[Dict] = None,
                  content: Optional[str] = None) -> List[Document]:
        """
        Parse file into string. If content is provided, use that instead of reading from file.

        Parameters:
            file (Path): The path to the markdown file.
            extra_info (Optional[Dict]): Additional metadata to include.
            content (Optional[str]): Content to use instead of reading from file.

        Returns:
            List[Document]: List of Document objects representing header-docs or a single Document object.
        """
        if self._read_as_single_doc:
            return self._load_data_as_single_doc(file, extra_info, content)
        else:
            return self._load_data_as_sections(file, extra_info, content)



import os
import shutil
from typing import Optional, Sequence

from llama_index import Document, ServiceContext, StorageContext, VectorStoreIndex
from llama_index.schema import BaseNode
from llama_index.vector_stores.types import VectorStore

from autollm.utils.env_utils import on_rm_error
from autollm.utils.lancedb_vectorstore import LanceDBVectorStore
from autollm.utils.logging import logger


def import_vector_store_class(vector_store_class_name: str):
    """
    Imports a predefined vector store class by class name.

    Args:
    Returns:
        The imported VectorStore class.
    """
    module = __import__("llama_index.vector_stores", fromlist=[vector_store_class_name])
    class_ = getattr(module, vector_store_class_name)
    return class_


class AutoVectorStoreIndex:
    """AutoVectorStoreIndex lets you dynamically initialize any Vector Store index based on the vector store
    class name and additional parameters.
    """

    @staticmethod
    def from_defaults(
            vector_store_type: str = "LanceDBVectorStore",
            lancedb_uri: str = None,
            lancedb_table_name: str = "vectors",
            lancedb_api_key: Optional[str] = None,
            lancedb_region: Optional[str] = None,
            use_async: bool = False,
            documents: Optional[Sequence[Document]] = None,
            nodes: Optional[Sequence[BaseNode]] = None,
            service_context: Optional[ServiceContext] = None,
            exist_ok: bool = False,
            overwrite_existing: bool = False,
            **kwargs) -> VectorStoreIndex:
        """
        Initializes a Vector Store index from Vector Store type and additional parameters. Handles lancedb
        path and document management according to specified behaviors.

        Parameters:
            vector_store_type (str): The class name of the vector store.
            lancedb_uri (str): The URI for the LanceDB vector store.
            lancedb_table_name (str): The table name for the LanceDB vector store.
            lancedb_api_key (Optional[str]): The API key for the LanceDB CLOUD vector store.
            lancedb_region (Optional[str]): The region for the LanceDB CLOUD vector store.
            use_async (bool): Flag to use async embedding. (Only supported for SimpleVectorStore)
            documents (Optional[Sequence[Document]]): Documents to initialize the vector store index from.
            service_context (Optional[ServiceContext]): Service context for initialization.
            exist_ok (bool): If True, allows adding to an existing database.
            overwrite_existing (bool): If True, allows overwriting an existing database.
            **kwargs: Additional parameters for initialization.

        Returns:
            VectorStoreIndex: The initialized Vector Store index instance.

        Raises:
            ValueError: For invalid parameter combinations or missing information.
        """
        if documents is None and nodes is None and vector_store_type == "SimpleVectorStore":
            raise ValueError("documents or nodes must be provided for SimpleVectorStore")

        if documents is not None and nodes is not None:
            raise ValueError("documents and nodes cannot be provided at the same time")

        if use_async and vector_store_type != "SimpleVectorStore":
            use_async = False
            logger.warning("use_async is only supported for SimpleVectorStore. Ignoring use_async.")

        # Initialize vector store
        VectorStoreClass = import_vector_store_class(vector_store_type)

        # If LanceDBVectorStore, use lancedb_uri and lancedb_table_name
        if vector_store_type == "LanceDBVectorStore":
            lancedb_uri = AutoVectorStoreIndex._validate_and_setup_lancedb_uri(
                lancedb_uri=lancedb_uri,
                documents=documents,
                exist_ok=exist_ok,
                overwrite_existing=overwrite_existing)

            vector_store = LanceDBVectorStore(
                uri=lancedb_uri,
                table_name=lancedb_table_name,
                api_key=lancedb_api_key,
                region=lancedb_region,
                **kwargs)

        else:
            vector_store = VectorStoreClass(**kwargs)

        # Initialize vector store index from existing vector store
        if documents is None and nodes is None:
            index = VectorStoreIndex.from_vector_store(
                vector_store=vector_store, service_context=service_context)
            return index

        # Initialize vector store index from documents or nodes
        index = AutoVectorStoreIndex._create_index(
            documents=documents,
            nodes=nodes,
            vector_store=vector_store,
            service_context=service_context,
            use_async=use_async,
            show_progress=True)

        return index

    @staticmethod
    def _validate_and_setup_lancedb_uri(lancedb_uri, documents, exist_ok, overwrite_existing):
        """
        Validates and sets up the lancedb_uri based on the given parameters.

        Parameters:
            lancedb_uri (str): The URI for the LanceDB vector store.
            documents (Sequence[Document]): Documents to initialize the vector store index from.
            exist_ok (bool): Flag to allow adding to an existing database.
            overwrite_existing (bool): Flag to allow overwriting an existing database.

        Returns:
            str: The validated and potentially modified lancedb_uri.
        """
        default_lancedb_uri = "./lancedb/db"

        # Scenario 0: Handle no lancedb uri and no documents provided
        if not documents and not lancedb_uri:
            raise ValueError(
                "A lancedb uri is required to connect to a database. Please provide a lancedb uri.")

        # Scenario 1: Handle lancedb_uri given but no documents provided
        if not documents and lancedb_uri:
            # Check if the database exists
            db_exists = os.path.exists(lancedb_uri)
            if not db_exists:
                raise ValueError(
                    f"No existing database found at {lancedb_uri}. Please provide a valid lancedb uri.")

        # Scenario 2: Handle no lancedb uri but documents provided
        if documents and not lancedb_uri:
            lancedb_uri = default_lancedb_uri
            lancedb_uri = AutoVectorStoreIndex._increment_lancedb_uri(lancedb_uri)
            logger.info(
                f"A new database is being created at {lancedb_uri}. Please provide a lancedb path to use an existing database."
            )

        # Scenario 3: Handle lancedb uri given and documents provided
        if documents and lancedb_uri:
            db_exists = os.path.exists(lancedb_uri)
            if exist_ok and overwrite_existing:
                if db_exists:
                    shutil.rmtree(lancedb_uri, onerror=on_rm_error)
                    logger.info(f"Overwriting existing database at {lancedb_uri}.")
            elif not exist_ok and overwrite_existing:
                raise ValueError("Cannot overwrite existing database without exist_ok set to True.")
            elif db_exists:
                if not exist_ok:
                    lancedb_uri = AutoVectorStoreIndex._increment_lancedb_uri(lancedb_uri)
                    logger.info(f"Existing database found. Creating a new database at {lancedb_uri}.")
                    logger.info(
                        "Please use exist_ok=True to add to the existing database and overwrite_existing=True to overwrite the existing database."
                    )
                else:
                    logger.info(f"Adding documents to existing database at {lancedb_uri}.")

        return lancedb_uri

    @staticmethod
    def _increment_lancedb_uri(base_uri: str) -> str:
        """Increment the lancedb uri to create a new database."""
        i = 1

        while os.path.exists(f"{base_uri}_{i}"):
            # check if the directory at path is empty. if so, use this path
            if not os.listdir(f"{base_uri}_{i}"):
                break
            i += 1
        return f"{base_uri}_{i}"

    @staticmethod
    def _create_index(
            documents: Optional[Sequence[Document]] = None,
            nodes: Optional[Sequence[BaseNode]] = None,
            vector_store: Optional[VectorStore] = None,
            service_context: Optional[ServiceContext] = None,
            use_async: Optional[bool] = False,
            show_progress: Optional[bool] = True):
        """
        Sets up the index from documents or nodes.

        Parameters:
            documents (Sequence[Document]): Documents to initialize the vector store index from.
            nodes (Sequence[BaseNode]): Nodes to initialize the vector store index from.
            vector_store (VectorStore): The vector store to initialize the index from.
            service_context (ServiceContext): Service context for initialization.
            use_async (bool): Flag to use async embedding.
            show_progress (bool): Flag to show progress.
        """
        if documents is None and nodes is None:
            raise ValueError("documents or nodes must be provided")

        if documents and nodes:
            raise ValueError("documents and nodes cannot be provided at the same time")

        storage_context = StorageContext.from_defaults(vector_store=vector_store)

        if documents is not None:
            index = VectorStoreIndex.from_documents(
                documents=documents,
                storage_context=storage_context,
                service_context=service_context,
                use_async=use_async,
                show_progress=show_progress)
        else:
            index = VectorStoreIndex(
                nodes=nodes,
                storage_context=storage_context,
                service_context=service_context,
                use_async=use_async,
                show_progress=show_progress)

        return index



from typing import Optional

from llama_index.llms import LiteLLM
from llama_index.llms.base import BaseLLM


class AutoLiteLLM:
    """AutoLiteLLM lets you dynamically initialize any LLM based on the llm class name and additional
    parameters.
    """

    @staticmethod
    def from_defaults(
            model: str = "gpt-3.5-turbo",
            max_tokens: Optional[int] = 256,
            temperature: float = 0.1,
            system_prompt: Optional[str] = None,
            api_base: Optional[str] = None) -> BaseLLM:
        """
        Create any LLM by model name. Check https://docs.litellm.ai/docs/providers for a list of
        supported models.

        If an argument is specified, then use the argument value provided for that
        parameter. If an argument is not specified, then use the default value.

        Parameters:
            model: Name of the LLM model to be initialized. Check
        https://docs.litellm.ai/docs/providers for a list of supported models.
            max_tokens: The maximum number of tokens to generate by the LLM.
            temperature: The temperature to use when sampling from the distribution.
            system_prompt: The system prompt to use for the LLM.
            api_base: The API base URL to use for the LLM.

        Returns:
            LLM: The initialized LiteLLM instance for given model name and parameter set.
        """

        return LiteLLM(
            model=model,
            max_tokens=max_tokens,
            temperature=temperature,
            system_prompt=system_prompt,
            api_base=api_base)



import hashlib
from pathlib import Path
from typing import List, Sequence, Tuple

from llama_index.schema import Document

from autollm.utils.logging import logger


def get_md5(file_path: Path) -> str:
    """
    Compute the MD5 hash of a file.

    Parameters:
        file_path (Path): The path to the file.

    Returns:
        str: The MD5 hash of the file.
    """
    hasher = hashlib.md5()
    with open(file_path, 'rb') as f:
        for chunk in iter(lambda: f.read(4096), b''):
            hasher.update(chunk)
    return hasher.hexdigest()


# TODO: add vs type
def check_for_changes(documents: Sequence[Document], vs) -> Tuple[Sequence[Document], List[str]]:
    """
    Check for file changes based on their hashes.

    Parameters:
        documents (Sequence[Document]): List of documents to check for changes.
        vs: The vector store to check for changes in.

    Returns:
        changed_documents (Sequence[Document]): List of documents that have changed.

        deleted_document_ids (List[str]): List of document ids that are deleted in local but present in vector store.
    """
    last_hashes, original_file_names, document_ids = vs.get_document_infos()
    deleted_document_ids = set(document_ids)

    changed_documents = []
    deleted_document_ids = []

    for doc in documents:
        file_path = str(Path(doc.metadata['original_file_path']))
        current_hash = get_md5(Path(file_path))

        # Add
        if file_path not in original_file_names:
            changed_documents.append(doc)
        # Update
        elif current_hash not in last_hashes:
            changed_documents.append(doc)
        else:
            # remove from deleted set
            deleted_document_ids.remove(doc.id_)

    deleted_document_ids = list(deleted_document_ids)

    logger.info(f'Found {len(changed_documents)} changed documents.')
    logger.info(f'Found {len(deleted_document_ids)} locally deleted documents still present in vector store.')

    return changed_documents, deleted_document_ids



from llama_index import Document, VectorStoreIndex
from llama_index.query_engine import BaseQueryEngine

from autollm.auto.vector_store_index import AutoVectorStoreIndex


def test_auto_vector_store():
    documents = [Document.example()]

    vector_store_index = AutoVectorStoreIndex.from_defaults(
        vector_store_type="SimpleVectorStore", documents=documents)

    # Check if the vector_store_index is an instance of VectorStoreIndex
    assert isinstance(vector_store_index, VectorStoreIndex)

    query_engine = vector_store_index.as_query_engine()

    # Check if the query_engine is an instance of BaseQueryEngine
    assert isinstance(query_engine, BaseQueryEngine)



DEFAULT_VECTORE_STORE_TYPE = 'pinecone'
DEFAULT_INDEX_NAME = 'quickstart'
DEFAULT_LLM_CLASS_NAME = 'OpenAI'

DEFAULT_RELATIVE_DOCS_PATH = 'docs'

DEFAULT_CHUNK_SIZE = 1024
DEFAULT_CHUNK_OVERLAP = 20
DEFAULT_CONTEXT_WINDOW = 4096

DEFAULT_MAX_TOKENS = 1024
DEFAULT_OPENAI_MODEL = 'gpt-3.5-turbo'
DEFAULT_PALM_MODEL = 'models/text-bison-001'
DEFAULT_ANYSCALE_MODEL = 'meta-llama/Llama-2-70b-chat-hf'

WEBPAGE_READER_TIMEOUT = 3



import logging
from typing import Dict, Optional, Sequence

from llama_index import Document
from llama_index.indices.query.base import BaseQueryEngine

from autollm.auto.query_engine import AutoQueryEngine
from autollm.utils.env_utils import load_config_and_dotenv

logging.basicConfig(level=logging.INFO)

STREAMING_CHUNK_SIZE = 16


def load_config_and_initialize_engines(
        config_file_path: str,
        env_file_path: str = None,
        documents: Optional[Sequence[Document]] = None) -> Dict[str, BaseQueryEngine]:
    """
    Load the YAML configuration file and optionally load environment variables from a .env file. Initialize
    query engines based on the config.

    Parameters:
        config_file_path (str): Path to the YAML configuration file.
        env_file_path (str): Path to the .env file.
        documents (Sequence[Document]): Sequence of llama_index.Document instances.

    Returns:
        dict: The configuration mapping (task name -> query engine))
    """

    config = load_config_and_dotenv(config_file_path, env_file_path)

    # Initialize query engines based on the config
    query_engines = {}
    for task_params in config['tasks']:
        task_name = task_params.pop('name')
        query_engines[task_name] = AutoQueryEngine.from_defaults(documents=documents, **task_params)

    return query_engines


def stream_text_data(text_data: str, chunk_size: int = STREAMING_CHUNK_SIZE):
    start = 0
    end = chunk_size
    while start < len(text_data):
        yield text_data[start:end]
        start = end
        end += chunk_size



from autollm.utils.templates import QUERY_PROMPT_TEMPLATE, SYSTEM_PROMPT


def set_default_prompt_template():
    """
    Sets the default prompt templates for the query engine.

    Returns:
        SystemPrompt (str): The default system prompt for the query engine.
        QueryPromptTemplate: The default query prompt template for the query engine.
    """

    return SYSTEM_PROMPT, QUERY_PROMPT_TEMPLATE



from llama_index import Document, ServiceContext, VectorStoreIndex

from autollm.auto.service_context import AutoServiceContext


def test_auto_service_context():
    document = Document.example()

    service_context = AutoServiceContext.from_defaults(enable_cost_calculator=True)

    # Check if the service_context is an instance of ServiceContext
    assert isinstance(service_context, ServiceContext)

    index = VectorStoreIndex.from_documents(documents=[document], service_context=service_context)

    query_engine = index.as_query_engine()

    response = query_engine.query("What is the meaning of life?")

    # Check if the response is not None
    assert response.response is not None

    # Check if the cost calculating handler is working
    cost_caltulator = service_context.callback_manager.handlers[0]
    assert cost_caltulator.total_llm_token_cost > 0



# Description: Templates for the system and user prompts.

SYSTEM_PROMPT = '''
Your purpose is to help users find the most relevant and accurate answers to their questions based on the documents you have access to.
You can answer questions based on the information available in the documents.
Your answers should be accurate, and directly related to the query.
When answering the questions, mostly rely on the info in documents.
'''

QUERY_PROMPT_TEMPLATE = '''
The document information is below.
---------------------
{context_str}
---------------------
Using the document information and mostly relying on it,
answer the query.
Query: {query_str}
Answer:
'''

REFINE_PROMPT_TEMPLATE = '''
The original query is as follows: {query_str}
You have provided an existing answer: {existing_answer}
Here is the new context:
------------
{context_msg}
------------
Given the new context, refine the original answer to better
answer the query (only if necessary).
Refined Answer:
'''



import shutil
from pathlib import Path
from typing import List, Optional, Sequence

from llama_index.readers.file.base import SimpleDirectoryReader
from llama_index.schema import Document

from autollm.utils.env_utils import on_rm_error
from autollm.utils.git_utils import clone_or_pull_repository
from autollm.utils.logging import logger
from autollm.utils.markdown_reader import MarkdownReader
from autollm.utils.pdf_reader import LangchainPDFReader
from autollm.utils.webpage_reader import WebPageReader
from autollm.utils.website_reader import WebSiteReader


def read_files_as_documents(
        input_dir: Optional[str] = None,
        input_files: Optional[List] = None,
        exclude_hidden: bool = True,
        filename_as_id: bool = True,
        recursive: bool = True,
        required_exts: Optional[List[str]] = None,
        show_progress: bool = True,
        **kwargs) -> Sequence[Document]:
    """
    Process markdown files to extract documents using SimpleDirectoryReader.

    Parameters:
        input_dir (str): Path to the directory containing the markdown files.
        input_files (List): List of file paths.
        exclude_hidden (bool): Whether to exclude hidden files.
        filename_as_id (bool): Whether to use the filename as the document id.
        recursive (bool): Whether to recursively search for files in the input directory.
        required_exts (Optional[List[str]]): List of file extensions to be read. Defaults to all supported extensions.

    Returns:
        documents (Sequence[Document]): A sequence of Document objects.
    """
    # Configure file_extractor to use MarkdownReader for md files
    file_extractor = {
        ".md": MarkdownReader(read_as_single_doc=True),
        ".pdf": LangchainPDFReader(extract_images=False)
    }

    # Initialize SimpleDirectoryReader
    reader = SimpleDirectoryReader(
        input_dir=input_dir,
        exclude_hidden=exclude_hidden,
        file_extractor=file_extractor,
        input_files=input_files,
        filename_as_id=filename_as_id,
        recursive=recursive,
        required_exts=required_exts,
        **kwargs)

    logger.info(f"Reading files from {input_dir}..") if input_dir else logger.info(
        f"Reading files {input_files}..")

    # Read and process the documents
    documents = reader.load_data(show_progress=show_progress)

    logger.info(f"Found {len(documents)} 'document(s)'.")
    return documents


def read_github_repo_as_documents(
        git_repo_url: str,
        relative_folder_path: Optional[str] = None,
        required_exts: Optional[List[str]] = None) -> Sequence[Document]:
    """
    A document provider that fetches documents from a specific folder within a GitHub repository.

    Parameters:
        git_repo_url (str): The URL of the GitHub repository.
        relative_folder_path (str, optional): The relative path from the repo root to the folder containing documents.
        required_exts (Optional[List[str]]): List of required extensions.

    Returns:
        Sequence[Document]: A sequence of Document objects.
    """

    # Ensure the temp_dir directory exists
    temp_dir = Path("autollm/temp/")
    temp_dir.mkdir(parents=True, exist_ok=True)

    logger.info(f"Cloning github repo {git_repo_url} into temporary directory {temp_dir}..")

    try:
        # Clone or pull the GitHub repository to get the latest documents
        clone_or_pull_repository(git_repo_url, temp_dir)

        # Specify the path to the documents
        docs_path = temp_dir if relative_folder_path is None else (temp_dir / Path(relative_folder_path))

        # Read and process the documents
        documents = read_files_as_documents(input_dir=str(docs_path), required_exts=required_exts)
        # Logging (assuming logger is configured)
        logger.info(f"Operations complete, deleting temporary directory {temp_dir}..")
    finally:
        # Delete the temporary directory
        shutil.rmtree(temp_dir, onerror=on_rm_error)

    return documents


def read_website_as_documents(
        parent_url: Optional[str] = None,
        sitemap_url: Optional[str] = None,
        include_filter_str: Optional[str] = None,
        exclude_filter_str: Optional[str] = None) -> List[Document]:
    """
    Read documents from a website or a sitemap.

    Parameters:
        parent_url (str, optional): The starting URL from which to scrape documents.
        sitemap_url (str, optional): The URL of the sitemap to process.
        include_filter_str (str, optional): Filter string to include certain URLs.
        exclude_filter_str (str, optional): Filter string to exclude certain URLs.

    Returns:
        List[Document]: A list of Document objects containing content and metadata.

    Raises:
        ValueError: If neither parent_url nor sitemap_url is provided, or if both are provided.
    """
    if (parent_url is None and sitemap_url is None) or (parent_url is not None and sitemap_url is not None):
        raise ValueError("Please provide either parent_url or sitemap_url, not both or none.")

    reader = WebSiteReader()
    if parent_url:
        documents = reader.load_data(
            parent_url=parent_url,
            include_filter_str=include_filter_str,
            exclude_filter_str=exclude_filter_str)
    else:
        documents = reader.load_data(
            sitemap_url=sitemap_url,
            include_filter_str=include_filter_str,
            exclude_filter_str=exclude_filter_str)

    return documents


def read_webpage_as_documents(url: str) -> List[Document]:
    """
    Read documents from a single webpage URL using the WebPageReader.

    Parameters:
        url (str): The URL of the web page to read.

    Returns:
        List[Document]: A list of Document objects containing content and metadata from the web page.
    """
    reader = WebPageReader()
    documents = reader.load_data(url)
    return documents



from typing import Optional, Sequence

from fastapi import FastAPI, HTTPException
from fastapi.responses import StreamingResponse
from llama_index import Document
from llama_index.indices.query.base import BaseQueryEngine
from pydantic import BaseModel, Field

from autollm.serve.docs import description, openapi_url, tags_metadata, terms_of_service, title, version
from autollm.serve.utils import load_config_and_initialize_engines, stream_text_data


class FromConfigQueryPayload(BaseModel):
    task: str = Field(..., description="Task to execute")
    user_query: str = Field(..., description="User's query")
    streaming: Optional[bool] = Field(False, description="Flag to enable streaming of response")


class FromEngineQueryPayload(BaseModel):
    user_query: str = Field(..., description="User's query")
    streaming: Optional[bool] = Field(False, description="Flag to enable streaming of response")


class AutoFastAPI:
    """Creates an FastAPI instance from config.yaml or Llama-Index query engine."""

    @staticmethod
    def from_config(
            config_file_path: Optional[str] = None,
            env_file_path: Optional[str] = None,
            task_name_to_query_engine: Optional[dict] = None,
            documents: Optional[Sequence[Document]] = None,
            api_title: str = None,
            api_description: str = None,
            api_version: str = None,
            api_term_of_service: str = None) -> FastAPI:
        """
        Create an FastAPI instance from config.yaml and optionally a .env file. The app has a single endpoint
        /query that takes a QueryPayload and returns a QueryResponse.

        ```python
        from autollm.auto.fastapi_app import QueryPayload, AutoFastAPI
        import uvicorn
        import requests

        # Start the server using from_config class method
        app = AutoFastAPI.from_config(config.yaml", ".env")
        uvicorn.run(app, host="0.0.0.0", port=8000)

        # Post request to the server
        data = {
          "task": "task1",
          "user_query": "why so serious?"
        }
        response = requests.post('http://0.0.0.0:8000/query', json=data)
        print(response.json())
        ```

        Parameters:
            config_file_path (str): Path to the YAML configuration file.
            env_file_path (str): Path to the .env file.
            task_name_to_query_engine (dict): Dictionary mapping task names to query engines.
            documents (Sequence[Document]): Sequence of llama_index.Document instances.
            api_title (str): Title of the API.
            api_description (str): Description of the API.
            api_version (str): Version of the API.
            api_term_of_service (str): Term of service of the API.

        Returns:
            FastAPI: The initialized FastAPI instance.
        """

        if task_name_to_query_engine is None and config_file_path is None:
            raise ValueError("Either config_file_path or task_name_to_query_engine must be provided")

        if task_name_to_query_engine is not None and config_file_path is not None:
            raise ValueError("Only one of config_file_path or task_name_to_query_engine must be provided")

        if task_name_to_query_engine is not None and not isinstance(task_name_to_query_engine, dict):
            raise ValueError("task_name_to_query_engine must be a dictionary")

        app = FastAPI(
            title=title if api_title is None else api_title,
            description=description if api_description is None else api_description,
            version=version if api_version is None else api_version,
            openapi_url=openapi_url,
            terms_of_service=terms_of_service if api_term_of_service is None else api_term_of_service,
            openapi_tags=tags_metadata,
        )

        if config_file_path is not None:
            task_name_to_query_engine = load_config_and_initialize_engines(
                config_file_path, env_file_path, documents)

        @app.post("/query")
        async def query(payload: FromConfigQueryPayload):
            task = payload.task
            user_query = payload.user_query

            if task not in task_name_to_query_engine:
                raise HTTPException(status_code=400, detail="Invalid task name")

            # Use the appropriate query engine for the task
            query_engine: BaseQueryEngine = task_name_to_query_engine[task]
            response = query_engine.query(user_query)

            # Check if the response should be streamed
            if payload.streaming:
                return StreamingResponse(stream_text_data(response.response))

            return response.response

        return app

    @staticmethod
    def from_query_engine(
            query_engine: BaseQueryEngine,
            api_title: str = None,
            api_description: str = None,
            api_version: str = None,
            api_term_of_service: str = None) -> FastAPI:
        """
        Create an FastAPI instance from a llama-index query engine.

        ```python
        from autollm.auto.fastapi_app import QueryPayload, AutoFastAPI
        import uvicorn
        import requests

        # Start the server using from_query_engine class method
        app = AutoFastAPI.from_query_engine(query_engine)
        uvicorn.run(app, host="0.0.0.0", port=8000)

        # Post request to the server
        data = {
           "user_query": "why so serious?"
        }
        response = requests.post('http://0.0.0.0:8000/query', json=data)

        print(response.json())
        ```

        Parameters:
            query_engine (BaseQueryEngine): Query engine.
            api_title (str): Title of the API.
            api_description (str): Description of the API.
            api_version (str): Version of the API.
            api_term_of_service (str): Term of service of the API.

        Returns:
            FastAPI: The initialized FastAPI instance.
        """

        if not isinstance(query_engine, BaseQueryEngine):
            raise ValueError("query_engine must be a llama_index query engine")

        app = FastAPI(
            title=title if api_title is None else api_title,
            description=description if api_description is None else api_description,
            version=version if api_version is None else api_version,
            openapi_url=openapi_url,
            terms_of_service=terms_of_service if api_term_of_service is None else api_term_of_service,
            openapi_tags=tags_metadata,
        )

        @app.post("/query")
        async def query(payload: FromEngineQueryPayload):
            user_query = payload.user_query

            response = query_engine.query(user_query)

            # Check if the response should be streamed
            if payload.streaming:
                return StreamingResponse(stream_text_data(response.response))

            return response.response

        return app



import os
import re

import setuptools


def get_requirements(req_path: str):
    with open(req_path, encoding='utf8') as f:
        return f.read().splitlines()


INSTALL_REQUIRES = get_requirements("requirements.txt")
DEV_REQUIREMETNS = get_requirements("dev-requirements.txt")
READERS_REQUIREMENTS = get_requirements("readers-requirements.txt")


def get_long_description():
    base_dir = os.path.abspath(os.path.dirname(__file__))
    with open(os.path.join(base_dir, 'README.md'), encoding='utf-8') as f:
        return f.read()


def get_version():
    current_dir = os.path.abspath(os.path.dirname(__file__))
    version_file = os.path.join(current_dir, 'autollm', '__init__.py')
    with open(version_file, encoding='utf-8') as f:
        return re.search(r'^__version__ = [\'"]([^\'"]*)[\'"]', f.read(), re.M).group(1)


def get_author():
    current_dir = os.path.abspath(os.path.dirname(__file__))
    init_file = os.path.join(current_dir, 'autollm', '__init__.py')
    with open(init_file, encoding='utf-8') as f:
        return re.search(r'^__author__ = [\'"]([^\'"]*)[\'"]', f.read(), re.M).group(1)


def get_license():
    current_dir = os.path.abspath(os.path.dirname(__file__))
    init_file = os.path.join(current_dir, 'autollm', '__init__.py')
    with open(init_file, encoding='utf-8') as f:
        return re.search(r'^__license__ = [\'"]([^\'"]*)[\'"]', f.read(), re.M).group(1)


setuptools.setup(
    name='autollm',
    version=get_version(),
    author=get_author(),
    author_email='support@safevideo.ai',
    license=get_license(),
    description="Ship RAG based LLM Web API's, in seconds.",
    long_description=get_long_description(),
    long_description_content_type='text/markdown',
    url='https://github.com/safevideo/autollm',
    packages=setuptools.find_packages(exclude=["tests", "examples"]),
    install_requires=INSTALL_REQUIRES,
    extras_require={
        'dev': DEV_REQUIREMETNS,
        'readers': READERS_REQUIREMENTS
    },
    python_requires='>=3.8',
    classifiers=[
        'Intended Audience :: Developers', 'Intended Audience :: Information Technology',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: GNU Affero General Public License v3',
        'Operating System :: MacOS :: MacOS X', 'Operating System :: Microsoft :: Windows',
        'Operating System :: POSIX', 'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9', 'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11', 'Topic :: Software Development',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'Topic :: Internet :: WWW/HTTP :: HTTP Servers'
    ],
)



from typing import Any

from litellm import aembedding as lite_aembedding
from litellm import embedding as lite_embedding
from llama_index.bridge.pydantic import Field
from llama_index.embeddings.base import BaseEmbedding, Embedding


class AutoEmbedding(BaseEmbedding):
    """
    Custom embedding class for flexible and efficient text embedding.

    This class interfaces with the LiteLLM library to use its embedding functionality, making it compatible
    with a wide range of LLM models.
    """

    # Define the model attribute using Pydantic's Field
    model: str = Field(default="text-embedding-ada-002", description="The name of the embedding model.")

    def __init__(self, model: str, **kwargs: Any) -> None:
        """
        Initialize the AutoEmbedding with a specific model.

        Args:
            model (str): ID of the embedding model to use.
            **kwargs (Any): Additional keyword arguments.
        """
        super().__init__(**kwargs)
        self.model = model  # Set the model ID for embedding

    def _get_query_embedding(self, query: str) -> Embedding:
        """
        Synchronously get the embedding for a query string.

        Args:
            query (str): The query text to embed.

        Returns:
            Embedding: The embedding vector.
        """
        response = lite_embedding(model=self.model, input=[query])
        return self._parse_embedding_response(response)

    async def _aget_query_embedding(self, query: str) -> Embedding:
        """
        Asynchronously get the embedding for a query string.

        Args:
            query (str): The query text to embed.

        Returns:
            Embedding: The embedding vector.
        """
        response = await lite_aembedding(model=self.model, input=[query])
        return self._parse_embedding_response(response)

    def _get_text_embedding(self, text: str) -> Embedding:
        """
        Synchronously get the embedding for a text string.

        Args:
            text (str): The text to embed.

        Returns:
            Embedding: The embedding vector.
        """
        return self._get_query_embedding(text)

    async def _aget_text_embedding(self, text: str) -> Embedding:
        """
        Asynchronously get the embedding for a text string.

        Args:
            text (str): The text to embed.

        Returns:
            Embedding: The embedding vector.
        """
        return await self._aget_query_embedding(text)

    def _parse_embedding_response(self, response):
        """
        Parse the embedding response from LiteLLM and extract the embedding data.

        Args:
            response: The response object from LiteLLM's embedding function.

        Returns:
            List[float]: The extracted embedding list.
        """
        try:
            if 'data' in response and len(response['data']) > 0 and 'embedding' in response['data'][0]:
                return response['data'][0]['embedding']
            else:
                raise ValueError("Invalid response structure from embedding function.")
        except (TypeError, KeyError, IndexError) as e:
            # Handle any parsing errors
            raise ValueError(f"Error parsing embedding response: {e}")



import os
import stat
from pathlib import Path
from typing import Callable, Tuple

import yaml
from dotenv import load_dotenv


def find_dotenv_file(start_path: Path) -> Path:
    """Searches for the .env file from start_path moving upwards."""
    current_path = start_path
    while current_path != Path('/'):
        dotenv_path = current_path / '.env'
        if dotenv_path.exists():
            return dotenv_path
        current_path = current_path.parent
    raise OSError('Could not find a .env file.')


# Find and load .env file
# dotenv_path = find_dotenv_file(Path(__file__).parent)
# load_dotenv(dotenv_path)


def read_env_variable(variable_name: str, default_value: str = None) -> str:
    """Reads an environment variable, returning a default value if not found."""
    return os.getenv(variable_name, default_value)


def validate_environment_variables(required_vars: list) -> None:
    """Validates that all required environment variables are set, raising an error if any of them is not
    set.
    """
    for var in required_vars:
        if var not in os.environ:
            raise OSError(f'Environment variable {var} is not set')


def load_config_and_dotenv(config_file_path: str, env_file_path: str = None) -> dict:
    """
    Load the YAML configuration file and optionally load environment variables from a .env file.

    Parameters:
        config_file_path (str): Path to the YAML configuration file.
        env_file_path (str): Path to the .env file.

    Returns:
        dict: The configuration dictionary.
    """
    # Optionally load environment variables from a .env file
    if env_file_path:
        load_dotenv(dotenv_path=env_file_path)

    # Load the YAML configuration file
    with open(config_file_path) as f:
        config = yaml.safe_load(f)

    return config


# From http://stackoverflow.com/a/4829285/548792
def on_rm_error(func: Callable, path: str, exc_info: Tuple):
    """
    Error handler for `shutil.rmtree` to handle permission errors.

    Parameters:
        func (Callable): The function that raised the error.
        path (str): The path to the file or directory which couldn't be removed.
        exc_info (Tuple): Exception information returned by sys.exc_info().
    """
    os.chmod(path, stat.S_IWRITE)
    os.unlink(path)



# db_utils.py
from typing import Sequence

from llama_index import Document, StorageContext, VectorStoreIndex
from llama_index.vector_stores import PineconeVectorStore, QdrantVectorStore

from autollm.utils.env_utils import read_env_variable
from autollm.utils.logging import logger


def initialize_pinecone_index(
        index_name: str, dimension: int = 1536, metric: str = 'euclidean', pod_type: str = 'p1'):
    import pinecone

    # Read environment variables for Pinecone initialization
    api_key = read_env_variable('PINECONE_API_KEY')
    environment = read_env_variable('PINECONE_ENVIRONMENT')

    # Initialize Pinecone
    pinecone.init(api_key=api_key, environment=environment)
    pinecone.create_index(index_name, dimension=dimension, metric=metric, pod_type=pod_type)


def initialize_qdrant_index(index_name: str, size: int = 1536, distance: str = 'EUCLID'):
    """Initialize Qdrant index."""
    from qdrant_client import QdrantClient
    from qdrant_client.models import Distance, VectorParams

    # Initialize client
    url = read_env_variable('QDRANT_URL')
    api_key = read_env_variable('QDRANT_API_KEY')

    client = QdrantClient(url=url, api_key=api_key)

    # Convert string distance measure to Distance Enum equals to Distance.EUCLID
    distance = Distance[distance]

    # Create index
    client.recreate_collection(
        collection_name=index_name, vectors_config=VectorParams(size=size, distance=distance))


def connect_vectorstore(vector_store, **params):
    """Connect to an existing vector store."""
    import pinecone
    from qdrant_client import QdrantClient

    # Logic to connect to vector store based on the specific type of vector store
    if isinstance(vector_store, PineconeVectorStore):
        vector_store.pinecone_index = pinecone.Index(params['index_name'])
    elif isinstance(vector_store, QdrantVectorStore):
        vector_store.client = QdrantClient(url=params['url'], api_key=params['api_key'])
    # TODO: Add more elif conditions for other vector stores as needed


def update_vector_store_index(vector_store_index: VectorStoreIndex, documents: Sequence[Document]):
    """
    Update the vector store index with new documents.

    Parameters:
        vector_store_index: An instance of AutoVectorStoreIndex or any compatible vector store.
        documents (Sequence[Document]): List of documents to update.

    Returns:
        None
    """
    for document in documents:
        delete_documents_by_id(vector_store_index, [document.id_])
        vector_store_index.insert(document)


def overwrite_vectorindex(vector_store, documents: Sequence[Document]):
    """
    Overwrite the vector store index with new documents.

    Parameters:
        vector_store: An instance of AutoVectorStore or any compatible vector store.
        documents (Sequence[Document]): List of documents to overwrite.

    Returns:
        None
    """
    # Create storage context
    storage_context = StorageContext.from_defaults(vector_store=vector_store)

    # Create index, which will insert documents/vectors to vector store
    _ = VectorStoreIndex.from_documents(documents, storage_context=storage_context)


def delete_documents_by_id(vector_store_index: VectorStoreIndex, document_ids: Sequence[str]):
    """
    Delete documents from vector store by their ids.

    Parameters:
        vector_store_index: An instance of AutoVectorStoreIndex or any compatible vector store.
        document_ids (Sequence[str]): List of document ids to delete.

    Returns:
        None
    """
    # Check if there are any document IDs to delete.
    if not document_ids:
        return

    # Proceed with deletion.
    for document_id in document_ids:
        vector_store_index.delete_ref_doc(document_id, delete_from_docstore=True)


# TODO: refactor and update.
# def initialize_database(
#         documents: Sequence[Document], vector_store_class_name: str, **vector_store_params) -> None:
#     logger.info('Initializing vector store')

#     vector_store = AutoVectorStore.from_defaults(vector_store_class_name, **vector_store_params)

#     if vector_store_class_name == 'PineconeVectorStore':
#         initialize_pinecone_index(vector_store, **vector_store_params)
#     elif vector_store_class_name == 'QdrantVectorStore':
#         initialize_qdrant_index(vector_store, **vector_store_params)
#     # TODO: Add more elif conditions for other vector stores as needed

#     connect_vectorstore(vector_store, **vector_store_params)

#     logger.info('Updating vector store with documents')

#     update_vector_store_index(vector_store, documents)

#     logger.info('Vector database successfully initialized.')

# # TODO: refactor and update.
# def update_database(documents: Sequence[Document], vectore_store_type: str) -> None:
#     """
#     Update the vector database to synchronize it with the provided list of documents.

#     This function performs the following actions:
#     1. Updates or adds new documents in the vector database that match the input list.
#     2. Removes any documents from the vector database that are not present in the input list.

#     Parameters:
#         documents (Sequence[Document]): Complete set of documents that should exist in the vector database after the update.
#         vectore_store_type (str): Specifies the type of vector store to use (e.g., 'qdrant', 'pinecone'). Defaults to DEFAULT_VECTORE_STORE_TYPE.

#     Returns:
#         None

#     Note:
#         Ensure that the 'documents' list includes all documents that should remain in the database, as any missing items will be deleted.
#     """
#     logger.info('Updating vector store')

#     # Get changed document ids using the hash of the documents available in the vector store index item metadata
#     vector_store = AutoVectorStore.from_defaults(
#         vector_store_type=vectore_store_type, index_name=DEFAULT_INDEX_NAME)
#     changed_documents, deleted_document_ids = check_for_changes(documents, vector_store)

#     # Update the index with the changed documents
#     vector_store.update_vectorindex(changed_documents)
#     vector_store.delete_documents_by_id(deleted_document_ids)

#     logger.info('Vector database successfully updated.')
